# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.core.mail import send_mail
from django.db import models, transaction
from django.db.models import Q
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.forms import ValidationError
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from allauth.account.models import EmailAddress
from django_countries.fields import CountryField
from localflavor.generic.countries.sepa import IBAN_SEPA_COUNTRIES
from localflavor.generic.models import IBANField
from memoize import delete_memoized, memoize
from rolepermissions.checkers import has_role
from rolepermissions.roles import assign_role, clear_roles

from apps.challenge.models import QualificationActivity, Season
from apps.common import (
    DV_LANGUAGES,
    DV_LANGUAGES_WITH_DEFAULT,
    DV_SEASON_STATE_PLANNING,
    DV_STATE_CHOICES,
    DV_STATE_CHOICES_WITH_DEFAULT,
    STDGLYPHICON,
)
from apps.common.models import Address
from defivelo.roles import has_permission, user_cantons

from ..common.fields import ChoiceArrayField
from . import FORMATION_CHOICES, formation_short, get_new_username

USERSTATUS_UNDEF = 0
USERSTATUS_ACTIVE = 10
USERSTATUS_RESERVE = 20
USERSTATUS_INACTIVE = 30
USERSTATUS_ARCHIVE = 40
USERSTATUS_DELETED = 99

USERSTATUS_CHOICES = (
    (USERSTATUS_UNDEF, "---------"),
    (USERSTATUS_ACTIVE, _("Actif·ve")),
    (USERSTATUS_RESERVE, _("Réserve")),
    (USERSTATUS_INACTIVE, _("Inactif·ve")),
    (USERSTATUS_ARCHIVE, _("Archivé·e")),
    (USERSTATUS_DELETED, _("Supprimé·e")),
)

USERSTATUS_CHOICES_NORMAL = tuple([us for us in USERSTATUS_CHOICES if us[0] < 90])

MARITALSTATUS_UNDEF = 0
MARITALSTATUS_SINGLE = 10
MARITALSTATUS_MARRIED = 20
MARITALSTATUS_DIVORCED = 30
MARITALSTATUS_WIDOW = 40
MARITALSTATUS_PACS = 50

MARITALSTATUS_CHOICES = (
    (MARITALSTATUS_UNDEF, "---------"),
    (MARITALSTATUS_SINGLE, _("Célibataire")),
    (MARITALSTATUS_MARRIED, _("Marié·e")),
    (MARITALSTATUS_DIVORCED, _("Divorcé·e")),
    (MARITALSTATUS_WIDOW, _("Veu·f·ve")),
    (MARITALSTATUS_PACS, _("En partenariat enregistré")),
)

BAGSTATUS_NONE = 0
BAGSTATUS_LOAN = 10
BAGSTATUS_PAID = 20
BAGSTATUS_GIFT = 30

BAGSTATUS_CHOICES = (
    (BAGSTATUS_NONE, "---"),
    (BAGSTATUS_LOAN, _("En prêt")),
    (BAGSTATUS_PAID, _("Payé")),
    (BAGSTATUS_GIFT, _("Offert")),
)

PERSONAL_FIELDS = [
    "language",
    "languages_challenges",
    "natel",
    "phone",
    "birthdate",
    "address_street",
    "address_no",
    "address_zip",
    "address_city",
    "address_canton",
    "nationality",
    "work_permit",
    "tax_jurisdiction",
    "bank_name",
    "iban",
    "employee_code",
    "social_security",
    "marital_status",
    "status",
    "activity_cantons",
]

DV_PUBLIC_FIELDS = [
    "formation",
    "formation_firstdate",
    "formation_lastdate",
    "actor_for",
    "pedagogical_experience",
    "firstmed_course",
    "firstmed_course_comm",
    "bagstatus",
    "affiliation_canton",
]

DV_PRIVATE_FIELDS = ["comments", "updated_at"]

COLLABORATOR_FIELDS = [
    "natel",
    "status",
    "affiliation_canton",
    "activity_cantons",
    "language",
    "formation",
    "actor_for",
]

STD_PROFILE_FIELDS = PERSONAL_FIELDS + DV_PUBLIC_FIELDS + DV_PRIVATE_FIELDS


class ExistingUserProfileManager(models.Manager):
    def get_queryset(self):
        return (
            super(ExistingUserProfileManager, self)
            .get_queryset()
            .exclude(status=USERSTATUS_DELETED)
        )


class UserProfile(Address, models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        related_name="profile",
        primary_key=True,
        on_delete=models.CASCADE,
    )
    employee_code = models.CharField(_("Code salarié·e"), max_length=63, blank=True)
    language = models.CharField(
        _("Langue"), max_length=7, choices=DV_LANGUAGES_WITH_DEFAULT, blank=True
    )

    languages_challenges = ChoiceArrayField(
        models.CharField(
            max_length=2,
            choices=DV_LANGUAGES,
        ),
        verbose_name=_("Prêt à animer en"),
        default=list,
        blank=True,
    )

    birthdate = models.DateField(_("Date de naissance"), blank=True, null=True)
    nationality = CountryField(_("Nationalité"), default="CH")
    work_permit = models.CharField(_("Permis de travail"), max_length=255, blank=True)
    tax_jurisdiction = models.CharField(
        _("Lieu d’imposition"), max_length=511, blank=True
    )
    bank_name = models.CharField(_("Nom de la banque"), max_length=511, blank=True)
    iban = IBANField(
        _("Coordonnées bancaires (IBAN)"),
        include_countries=IBAN_SEPA_COUNTRIES,
        blank=True,
    )
    social_security = models.CharField(_("N° AVS"), max_length=16, blank=True)
    natel = models.CharField(max_length=17, blank=True)
    phone = models.CharField(_("Téléphone"), max_length=17, blank=True)
    affiliation_canton = models.CharField(
        _("Canton d’affiliation"),
        choices=DV_STATE_CHOICES_WITH_DEFAULT,
        max_length=2,
        blank=True,
    )

    activity_cantons = ChoiceArrayField(
        models.CharField(
            max_length=2,
            choices=DV_STATE_CHOICES,
        ),
        verbose_name=_("Défi Vélo mobile"),
        default=list,
        blank=True,
    )

    formation = models.CharField(
        _("Formation"), max_length=2, choices=FORMATION_CHOICES, blank=True
    )
    formation_firstdate = models.DateField(
        _("Date de la première formation"), blank=True, null=True
    )
    formation_lastdate = models.DateField(
        _("Date de la dernière formation"), blank=True, null=True
    )
    actor_for = models.ManyToManyField(
        QualificationActivity,
        verbose_name=_("Intervenant·e"),
        related_name="actor_for",
        limit_choices_to={"category": "C"},
        blank=True,
    )
    marital_status = models.PositiveSmallIntegerField(
        _("État civil"), choices=MARITALSTATUS_CHOICES, default=MARITALSTATUS_UNDEF
    )
    status = models.PositiveSmallIntegerField(
        _("Statut"), choices=USERSTATUS_CHOICES, default=USERSTATUS_ACTIVE
    )
    status_updatetime = models.DateTimeField(null=True, blank=True)
    pedagogical_experience = models.TextField(_("Expérience pédagogique"), blank=True)
    firstmed_course = models.BooleanField(_("Cours samaritains"), default=False)
    firstmed_course_comm = models.CharField(
        _("Cours samaritains (spécifier)"), max_length=255, blank=True
    )
    bagstatus = models.PositiveSmallIntegerField(
        _("Sac Défi Vélo"), choices=BAGSTATUS_CHOICES, default=BAGSTATUS_NONE
    )
    bagstatus_updatetime = models.DateTimeField(null=True, blank=True)
    comments = models.TextField(_("Remarques"), blank=True)

    updated_at = models.DateTimeField(auto_now=True)

    objects = models.Manager()
    objects_existing = ExistingUserProfileManager()

    @staticmethod
    def get_users_that_worked_in_cantons(
        cantons: list[str],
        additional_or: list[Q] | None = None,
        year: int | None = None,
    ):
        """
        Get all the users that worked at least once in the given states

        Users who participated in the qualifications
        (which took place in the given cantons), as:
        - leader (monitor 1)
        - helpers (monitor 2)
        - actor (meeting - Qualif C)
        """

        leader_filters = {
            f"qualifs_mon2__{k}": v
            for k, v in {"session__orga__address_canton__in": cantons}.items()
        }
        helper_filters = {
            f"qualifs_mon1__{k}": v
            for k, v in {"session__orga__address_canton__in": cantons}.items()
        }
        actor_filters = {
            f"qualifs_actor__{k}": v
            for k, v in {"session__orga__address_canton__in": cantons}.items()
        }

        if year is not None:
            leader_filters["qualifs_mon2__session__day__year"] = year
            helper_filters["qualifs_mon1__session__day__year"] = year
            actor_filters["qualifs_actor__session__day__year"] = year

        criteria = Q(**leader_filters) | Q(**helper_filters) | Q(**actor_filters)

        for q in additional_or if additional_or else []:
            criteria = criteria | q

        return (
            get_user_model()
            .objects.filter(criteria)
            .prefetch_related("profile")
            .distinct()
        )

    def save(self, *args, **kwargs):
        if self.activity_cantons:
            # Remove the affiliation canton from the activity_cantons
            try:
                self.activity_cantons.remove(self.affiliation_canton)
            except (ValueError, AttributeError):
                pass
        if self.languages_challenges:
            if not self.language:
                self.language = self.languages_challenges.pop(0)
            # Remove the main language from the languages_challenges
            try:
                self.languages_challenges.remove(self.language)
            except (ValueError, AttributeError):
                pass
        # Assign the "collaborator" role automatically for "with formation" and "actors"
        if not (
            has_role(self.user, "power_user")
            or has_role(self.user, "state_manager")
            or has_role(self.user, "coordinator")
        ):
            # self.actor is a cached property, so access the ManyToMany directly
            if self.formation or self.actor_for.exists():
                self.set_role("collaborator")
            else:
                # Remove the role
                self.set_role()

        self.reset_cache()
        super().save(*args, **kwargs)

    def reset_cache(self):
        delete_memoized(has_permission)
        delete_memoized(user_cantons, self.user)
        delete_memoized(self.get_seasons)

    def set_role(self, role_str=None):
        # Enforce that a user can only have one role at a time
        clear_roles(self.user)
        if role_str:
            assign_role(self.user, role_str)
        self.reset_cache()

    def set_statemanager_for(self, states=[]):
        for ums in self.user.managedstates.all():
            if ums.canton not in states:
                ums.delete()
            else:
                states.remove(ums.canton)
        for canton in states:
            UserManagedState.objects.get_or_create(user=self.user, canton=canton)
        self.reset_cache()

    def send_credentials(self, context, force=False):
        if self.can_login and not force:
            # Has credentials already
            raise ValidationError(
                _("A déjà des données de connexion"), code="has_login"
            )

        newpassword = get_user_model().objects.make_random_password()
        self.user.set_password(newpassword)
        self.user.is_active = True

        context["userprofile"] = self.user
        context["password"] = newpassword
        self.user.save()

        # This can raise exception, but that's good
        self.send_mail(
            (settings.EMAIL_SUBJECT_PREFIX + gettext("Accès à l’Intranet")),
            render_to_string("auth/email_user_send_credentials.txt", context),
        )

        # Create a validated email
        EmailAddress.objects.get_or_create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )

    @transaction.atomic
    def delete(self):
        self.user.is_active = False
        email_comment = _("Email avant suppression: %s") % self.user.email
        self.user.email = ""
        self.user.set_unusable_password()
        self.user.save()
        self.comments = (
            f"{self.comments}\n{email_comment}" if self.comments else email_comment
        )
        self.status = USERSTATUS_DELETED
        self.save()

    @cached_property
    def formation_full(self):
        if self.formation:
            return dict(FORMATION_CHOICES)[self.formation]
        return ""

    @cached_property
    def status_full(self):
        if self.status:
            return dict(USERSTATUS_CHOICES)[self.status]
        return ""

    @cached_property
    def marital_status_full(self):
        if self.marital_status:
            return dict(MARITALSTATUS_CHOICES)[self.marital_status]
        return ""

    def status_icon(self):
        icon = ""
        title = self.status_full
        if self.status == USERSTATUS_ACTIVE:
            icon = "star"
        elif self.status == USERSTATUS_RESERVE:
            icon = "star-empty"
        elif self.status == USERSTATUS_INACTIVE:
            icon = "hourglass"
        elif self.status == USERSTATUS_ARCHIVE:
            icon = "folder-close"
        elif self.status == USERSTATUS_DELETED:
            icon = "trash"
        if icon:
            return mark_safe(STDGLYPHICON.format(icon=icon, title=title))
        return ""

    def status_class(self):
        css_class = "default"
        if self.status == USERSTATUS_ACTIVE:
            css_class = "success"  # Green
        elif self.status == USERSTATUS_RESERVE:
            css_class = "warning"  # Orange
        elif self.status == USERSTATUS_INACTIVE:
            css_class = "danger"  # Red
        elif self.status == USERSTATUS_DELETED:
            css_class = "default disabled"  # Black
        return css_class

    @cached_property
    def age(self):
        today = timezone.now()
        return (
            today.year
            - self.birthdate.year
            - ((today.month, today.day) < (self.birthdate.month, self.birthdate.day))
        )

    @cached_property
    def iban_nice(self):
        if self.iban:
            value = self.iban
            # Code stolen from
            # https://django-localflavor.readthedocs.org/en/latest/_modules/localflavor/generic/forms/#IBANFormField.prepare_value
            grouping = 4
            value = value.upper().replace(" ", "").replace("-", "")
            return " ".join(
                value[i : i + grouping] for i in range(0, len(value), grouping)
            )
        return ""

    def formation_icon(self):
        return formation_short(self.formation)

    @cached_property
    def actor(self):
        return self.actor_for.exists()

    @cached_property
    def has_address(self):
        """
        Whether we have any address field set
        """
        address_fields = [
            f.name for f in Address._meta.get_fields() if f.name.startswith("address_")
        ]
        return any([getattr(self, field, False) != "" for field in address_fields])

    @property
    def is_paid_staff(self):
        """
        Whether a UserProfile is probably a paid staff
        """
        return self.affiliation_canton

    @cached_property
    def actor_inline(self):
        return " - ".join([smart_str(a) for a in self.actor_for.all()])

    def actor_icon(self):
        if self.actor_inline:
            return mark_safe(
                STDGLYPHICON.format(icon="sunglasses", title=self.actor_inline)
            )
        return ""

    @cached_property
    def bagstatus_full(self):
        if self.bagstatus:
            return dict(BAGSTATUS_CHOICES)[self.bagstatus]
        return ""

    def bagstatus_icon(self):
        icon = ""
        title = self.bagstatus_full
        if self.bagstatus == BAGSTATUS_NONE:
            icon = "unchecked"
        elif self.bagstatus == BAGSTATUS_LOAN:
            icon = "new-window"
        elif self.bagstatus == BAGSTATUS_PAID or self.bagstatus == BAGSTATUS_GIFT:
            icon = "check"
        if icon:
            return mark_safe(STDGLYPHICON.format(icon=icon, title=title))
        return ""

    @cached_property
    def language_verb(self):
        try:
            return [c[1] for c in DV_LANGUAGES if c[0] == self.language][0]
        except IndexError:
            return ""

    def access_level(self, textonly=True):
        icon = ""
        title = ""
        if self.can_login:
            title = _("A accès")
            icon = "user"

            if self.user.is_superuser:
                title = _("Administra·teur·trice")
                icon = "queen"
            elif self.user.groups.exists():
                if has_role(self.user, "power_user"):
                    title = _("Bureau de coordination")
                    icon = "king"
                elif has_role(self.user, "state_manager"):
                    title = _("Chargé·e de projet")
                    icon = "bishop"
                elif has_role(self.user, "coordinator"):
                    title = _("Coordina·teur·trice")
                    icon = "pawn"
                elif has_role(self.user, "collaborator"):
                    title = _("Collabora·teur·trice")
                    icon = "user"
        if title and textonly:
            return title
        if icon:
            return mark_safe(STDGLYPHICON.format(icon=icon, title=title))
        return ""

    @cached_property
    def access_level_icon(self):
        return self.access_level(False)

    @cached_property
    def access_level_text(self):
        return self.access_level(True)

    @cached_property
    def managed_cantons(self):
        return [m.canton for m in self.user.managedstates.all()]

    @cached_property
    def mailtolink(self):
        return "{name} <{email}>".format(
            name=self.user.get_full_name(), email=self.user.email
        )

    @cached_property
    def languages_challenges_text(self):
        languages = dict(DV_LANGUAGES)
        return ", ".join([str(languages[code]) for code in self.languages_challenges])

    @cached_property
    def can_login(self):
        return self.user.is_active and self.user.has_usable_password()

    @memoize()
    def get_seasons(self, raise_without_cantons=False):
        qs = Season.objects
        usercantons = []

        try:
            # Obtient les cantons gérés
            usercantons = user_cantons(self.user)
        except LookupError:
            if raise_without_cantons:
                raise PermissionDenied

        # Ajoute les cantons d’affiliation et mobiles
        if self.formation or self.actor:
            if self.affiliation_canton:
                usercantons += [self.affiliation_canton]
            if self.activity_cantons:
                usercantons += self.activity_cantons
            if not has_permission(self.user, "challenge_season_see_state_planning"):
                # PLANNING seasons are invisible for these
                qs = qs.exclude(
                    state__in=[
                        DV_SEASON_STATE_PLANNING,
                    ]
                )

        # Ajoute les cantons dans lesquels on a des établissements
        usercantons += [
            orga.address_canton for orga in self.user.managed_organizations.all()
        ]

        # Unique'ify, discard empty values
        usercantons = list(set(usercantons))

        # Au moins un canton en commun
        return qs.filter(cantons__overlap=usercantons) if usercantons else qs.none()

    @cached_property
    def deleted(self):
        return (
            self.status == USERSTATUS_DELETED
            and not self.user.is_active
            and not self.user.email
            and not self.user.has_usable_password()
        )

    def send_mail(self, subject: str, body: str):
        """
        Send an email to that user
        """
        if self.user.email:
            return send_mail(
                subject,
                body,
                settings.DEFAULT_FROM_EMAIL,
                [
                    self.mailtolink,
                ],
            )

    def __str__(self):
        return self.user.get_full_name()

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.id)

    class Meta:
        verbose_name = _("Profil")
        verbose_name_plural = _("Profils")


@receiver(pre_save, sender=settings.AUTH_USER_MODEL)
def User_pre_save(sender, **kwargs):
    if not kwargs["instance"].username:
        kwargs["instance"].username = get_new_username()
        # Mark new users as inactive, to not let them get a login
        kwargs["instance"].is_active = False


class UserManagedState(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="managedstates",
        limit_choices_to={"is_active": True},
        on_delete=models.CASCADE,
    )
    canton = models.CharField(_("Canton"), max_length=2, choices=DV_STATE_CHOICES)

    @property
    def canton_full(self):
        return [c[1] for c in DV_STATE_CHOICES if c[0] == self.address_canton][0]

    def __str__(self):
        return _("{name} est chargé·e de projet pour le canton {canton}").format(
            name=self.user.get_full_name(), canton=self.canton
        )

    class Meta:
        verbose_name = _("Canton géré")
        verbose_name_plural = _("Cantons gérés")
        unique_together = (("user", "canton"),)



import datetime

import factory

from apps.salary.models import MonthlyCantonalValidation, Timesheet
from apps.user.tests.factories import UserFactory


class TimesheetFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Timesheet

    user = factory.SubFactory(UserFactory)
    date = factory.LazyFunction(lambda: datetime.date.today())


class ValidatedTimesheetFactory(TimesheetFactory):
    validated_at = factory.LazyAttribute(
        lambda timesheet: timesheet.date + datetime.timedelta(days=1)
    )
    validated_by = factory.SubFactory(UserFactory)


class MonthlyCantonalValidationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MonthlyCantonalValidation

    date = factory.LazyFunction(lambda: datetime.date.today())



# Generated by Django 4.2.6 on 2025-11-13 13:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orga', '0008_auto_20200518_0322'),
    ]

    operations = [
        migrations.AlterField(
            model_name='organization',
            name='status',
            field=models.PositiveSmallIntegerField(choices=[(0, '---------'), (10, 'Actif·ve'), (30, 'Inactif·ve')], default=10, verbose_name='Statut'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='status',
            field=models.PositiveSmallIntegerField(
                choices=[(0, '---------'), (10, 'Actif'), (30, 'Inactif')], default=10,
                verbose_name='Statut'),
        ),
    ]



# Generated by Django 1.11.4 on 2017-08-24 10:12
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0037_auto_20170824_1109"),
    ]

    operations = [
        migrations.AlterField(
            model_name="qualification",
            name="activity_A",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="qualifs_A",
                to="challenge.QualificationActivity",
                verbose_name="Agilité",
            ),
        ),
        migrations.AlterField(
            model_name="qualificationactivity",
            name="category",
            field=models.CharField(
                blank=True,
                choices=[("A", "Agilité"), ("B", "Mécanique"), ("C", "Rencontre")],
                max_length=1,
                verbose_name="Catégorie",
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="cantons",
            field=models.CharField(
                choices=[
                    ("AR", "Appenzell Ausserrhoden"),
                    ("BE", "Berne"),
                    ("BS", "Basel-Stadt"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("VS-OW", "Haut-Valais"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("SG", "St. Gallen"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZH", "Zurich"),
                ],
                max_length=41,
                verbose_name="Cantons",
            ),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="session",
            name="timeslot",
            field=models.ForeignKey(
                to="challenge.SessionTimeSlot",
                blank=True,
                null=True,
                related_name="sessions",
                on_delete=models.CASCADE,
            ),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015,2020 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, time, timedelta

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import F, Q
from django.template.defaultfilters import date
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from simple_history.models import HistoricalRecords

from apps.common.models import Address
from apps.orga.models import ORGASTATUS_ACTIVE, Organization
from apps.salary.models import Timesheet
from apps.user import FORMATION_KEYS, FORMATION_M2

from .. import (
    CHOSEN_AS_ACTOR,
    CHOSEN_AS_HELPER,
    CHOSEN_AS_LEADER,
    CHOSEN_AS_NOT,
    CHOSEN_AS_REPLACEMENT,
    utils,
)

DEFAULT_SESSION_DURATION_HOURS = 3
DEFAULT_EARLY_MINUTES_FOR_HELPERS_MEETINGS = 60


class Session(Address, models.Model):
    created_on = models.DateTimeField(auto_now_add=True)

    # Time span
    day = models.DateField(_("Date"), blank=True)
    begin = models.TimeField(_("Début"), blank=True, null=True)
    duration = models.DurationField(
        _("Durée"), default=timedelta(hours=DEFAULT_SESSION_DURATION_HOURS)
    )
    visible = models.BooleanField(_("Visible pour les moniteur·trice·s"), default=False)
    orga = models.ForeignKey(
        Organization,
        verbose_name=_("Établissement"),
        related_name="sessions",
        limit_choices_to={
            "address_canton__isnull": False,
            "status__in": [ORGASTATUS_ACTIVE],
        },
        on_delete=models.CASCADE,
    )  # Don't delete orgas
    place = models.CharField(_("Lieu de la Qualif’"), max_length=512, blank=True)
    superleader = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Moniteur·trice + / Photographe"),
        related_name="sess_monplus",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    FALLBACK_CHOICES = (
        ("A", _("Programme déluge")),
        ("B", _("Annulation")),
        ("C", _("Report …")),
        ("D", _("Autre …")),
    )
    fallback_plan = models.CharField(
        _("Mauvais temps"), max_length=1, choices=FALLBACK_CHOICES, blank=True
    )
    helpers_time = models.TimeField(
        _("Heure de rendez-vous des moniteur·trice·s"), null=True, blank=True
    )
    helpers_place = models.CharField(
        _("Lieu rendez-vous moniteur·trice·s"), max_length=512, blank=True
    )
    apples = models.CharField(_("Pommes"), max_length=512, blank=True)
    bikes_concept = models.CharField(_("Logistique vélos"), max_length=512, blank=True)
    bikes_phone = models.CharField(_("N° de contact vélos"), max_length=13, blank=True)
    comments = models.TextField(_("Remarques"), blank=True)
    history = HistoricalRecords()

    class Meta:
        verbose_name = _("Session")
        verbose_name_plural = _("Sessions")
        ordering = ["day", "begin", "orga__name"]

    def get_related_timesheets(self):
        return Timesheet.objects.filter(
            (
                Q(user__qualifs_actor__in=self.qualifications.all())
                | Q(user__qualifs_mon2__in=self.qualifications.all())
                | Q(user__qualifs_mon1__in=self.qualifications.all())
            )
            & Q(date=self.day)
        )

    def has_related_timesheets(self):
        if self.pk:
            return self.get_related_timesheets().exists()
        return False

    @cached_property
    def has_availability_incoherences(self):
        for quali in self.qualifications.all():
            if quali.has_availability_incoherences:
                return True
        return False

    @property
    def errors(self):
        errors = []
        if not self.visible:
            errors.append(_("Visibilité"))
        if not self.begin or not self.duration:
            errors.append(_("Horaire"))
        if not self.fallback_plan:
            errors.append(_("Mauvais temps"))
        if not self.apples:
            errors.append(_("Pommes"))
        if not self.bikes_concept and not self.bikes_phone:
            errors.append(_("Logistique vélos"))
        if not self.address_city:
            errors.append(_("Emplacement"))
        # Check the qualifications
        qualiq = 0
        for quali in self.qualifications.all():
            qualiq += 1
            if quali.errors:
                errors.append(_("Qualif’ {name}").format(name=quali.name))
        if qualiq == 0:
            errors.append(_("Pas de Qualifs"))
        if errors:
            return mark_safe(
                "<br />".join(
                    [
                        '<span class="btn-warning btn-xs disabled">'
                        '  <span class="glyphicon glyphicon-warning-sign"></span>'
                        "  {error}"
                        "</span>".format(error=e)
                        for e in errors
                    ]
                )
            )

    @property
    def start_datetime(self):
        return datetime.combine(
            self.day, self.begin if self.begin is not None else time.min
        )

    @property
    def end_datetime(self):
        return self.start_datetime + self.duration

    @cached_property
    def fallback(self):
        if self.fallback_plan:
            return dict(self.FALLBACK_CHOICES)[self.fallback_plan]
        return ""

    @cached_property
    def end(self):
        if self.begin and self.duration:
            day = self.day if self.day else datetime.today()
            end = datetime.combine(day, self.begin) + self.duration
            return end.time()

    @cached_property
    def chosen_staff(self):
        return (
            self.availability_statuses.exclude(
                Q(availability="n") | Q(chosen_as=CHOSEN_AS_NOT),
            )
            .prefetch_related("helper", "helper__profile", "session")
            .order_by("-chosen_as", "-availability", "-helper__profile__formation")
        )

    @cached_property
    def replacement_staff(self):
        qualifs_pks = self.qualifications.values_list("id", flat=True)
        return (
            self.chosen_staff.filter(chosen_as=CHOSEN_AS_REPLACEMENT)
            .exclude(helper__qualifs_mon1__in=qualifs_pks)
            .exclude(helper__qualifs_mon2__in=qualifs_pks)
            .exclude(helper__qualifs_actor__in=qualifs_pks)
        )

    @cached_property
    def n_qualifications(self):
        return self.qualifications.count()

    def chosen_helpers(self):
        return self.chosen_staff.filter(
            chosen_as__in=[CHOSEN_AS_HELPER, CHOSEN_AS_LEADER],
            helper__profile__formation__in=FORMATION_KEYS,
        ).order_by(
            "-helper__profile__formation", "helper__first_name", "helper__last_name"
        )

    def chosen_helpers_M2(self):
        return self.chosen_helpers().filter(helper__profile__formation=FORMATION_M2)

    def helper_needs(self):
        # Return the number of helpers needed for that session.
        total_m1 = sum(q.n_helpers_enum.m1 for q in self.qualifications.all())
        total_m2 = sum(q.n_helpers_enum.m2 for q in self.qualifications.all())
        # Array index 1 is the number of M1 needed, index 2 the number of M2 needed.
        return [0, total_m1, total_m2]

    def chosen_actors(self):
        return self.chosen_staff.filter(chosen_as=CHOSEN_AS_ACTOR).exclude(
            helper__profile__actor_for__isnull=True
        )

    def actor_needs(self):
        return self.n_qualifications

    def user_assignment(self, user):
        # Check as what a user is assigned to that session
        for q in self.qualifications.all():
            if user == q.leader:
                return CHOSEN_AS_LEADER
            if user in q.helpers.all():
                return CHOSEN_AS_HELPER
            if user == q.actor:
                return CHOSEN_AS_ACTOR
        return None

    def n_quali_things(self, field):
        return sum([q for q in self.qualifications.values_list(field, flat=True) if q])

    @cached_property
    def n_bikes(self):
        return self.n_quali_things("n_bikes")

    @cached_property
    def n_helmets(self):
        return self.n_quali_things("n_helmets")

    @cached_property
    def n_participants(self):
        return self.n_quali_things("n_participants")

    def helpers_time_with_default(self):
        if self.helpers_time:
            return self.helpers_time
        if self.begin:
            # Compute a default value with regards to the start time
            helpers_time = date(
                (
                    datetime.combine(datetime.today(), self.begin)
                    - timedelta(minutes=DEFAULT_EARLY_MINUTES_FOR_HELPERS_MEETINGS)
                ).time(),
                settings.TIME_FORMAT,
            )
            return mark_safe("<em>{}</em>".format(helpers_time))
        return ""

    @cached_property
    def city(self):
        if self.address_city:
            return self.address_city
        elif self.orga.address_city:
            return self.orga.address_city

    @cached_property
    def season(self):
        from .season import Season

        season = Season.objects.filter(
            year=self.day.year,
            month_start__lte=self.day.month,
            n_months__gt=self.day.month - F("month_start"),
            cantons__contains=[self.orga.address_canton],
        ).first()

        if season:
            return season

        # Should only rarely happen, so afford an other request; look for Season objects in the past year, spanning to ours.
        for offset in [1, 2]:
            seasons = Season.objects.filter(
                year=self.day.year - offset,
                cantons__contains=[self.orga.address_canton],
            )
            for s in seasons.all():
                if s.begin <= self.day <= s.end:
                    return s

    @cached_property
    def short(self):
        return _("{place} {date}{time}").format(
            date=date(self.day, settings.DATE_FORMAT_SHORT),
            time=(
                "@" + date(self.begin, settings.TIME_FORMAT_SHORT) if self.begin else ""
            ),
            place=self.orga.name,
        )

    def clean(self):
        """Avoid creating two sessions at the same place, same day and same half of
        the day.
        """
        try:
            sessions = Session.sessions_in_same_half_day(
                self.orga, self.day, self.begin
            )
            if sessions.exclude(pk=self.pk).exists():
                raise ValidationError(
                    {
                        "day": _(
                            "Il y a déjà une session inscrite pour cet établissement à cette date."
                        )
                    }
                )
        except Session.orga.RelatedObjectDoesNotExist:
            # It's ok if the organisation is not yet selected
            pass

    @classmethod
    def sessions_in_same_half_day(cls, orga, day, begin):
        """Get the sessions for this orga, day, and at the same half of
        the day (beginning also before or after noon).
        """
        sessions = cls.objects.filter(orga_id=orga.id, day=day)
        if utils.is_morning(begin):
            sessions = sessions.filter(begin__lte=time(12, 00))
        else:
            sessions = sessions.filter(begin__gt=time(12, 00))

        return sessions

    def __str__(self):
        return (
            date(self.day, settings.DATE_FORMAT)
            + (" (%s)" % date(self.begin, settings.TIME_FORMAT) if self.begin else "")
            + (
                " - %s" % (self.orga.abbr if self.orga.abbr else self.orga.name)
                if self.orga
                else ""
            )
            + (
                " - %s"
                % (
                    self.address_city
                    if self.address_city
                    else (
                        self.orga.address_city
                        if (self.orga and self.orga.address_city)
                        else ""
                    )
                )
            )
        )

    def get_export_filename(self):
        """
        Same as __str__ but without spaces and parentheses
        """
        orga = (
            "_%s"
            % (self.orga.abbr if self.orga.abbr else self.orga.name).replace(" ", "-")
            if self.orga
            else ""
        )

        return (
            date(self.day, "Ymd")
            + ("_%s" % date(self.begin, "Hi") if self.begin else "")
            + orga
            + (
                "_%s"
                % (
                    self.address_city
                    if self.address_city
                    else (
                        self.orga.address_city
                        if (self.orga and self.orga.address_city)
                        else ""
                    )
                )
            )
        )



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2016 Didier Raboud <me+defivelo@odyx.org>
# Copyright (C) 2020 Didier Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .invoice import (
    InvoiceCreateView,
    InvoiceDetailView,
    InvoiceListView,
    InvoiceUpdateView,
    InvoiceYearlyListView,
    SeasonOrgaListView,
)
from .quali import QualiCreateView, QualiDeleteView, QualiUpdateView
from .season import (
    SeasonActorListView,
    SeasonAvailabilityReminderView,
    SeasonAvailabilityUpdateView,
    SeasonAvailabilityView,
    SeasonCreateView,
    SeasonDetailView,
    SeasonErrorsListView,
    SeasonExportView,
    SeasonHelperListView,
    SeasonListRedirect,
    SeasonListView,
    SeasonPersonalPlanningExportFeed,
    SeasonPersonalPlanningExportView,
    SeasonPlanningExportView,
    SeasonPlanningView,
    SeasonStaffChoiceUpdateView,
    SeasonToOpenView,
    SeasonToRunningView,
    SeasonUpdateView,
)
from .session import (
    SessionCreateView,
    SessionDeleteView,
    SessionDetailView,
    SessionExportView,
    SessionsListView,
    SessionStaffChoiceView,
    SessionUpdateView,
)
from .settings import (
    AnnualStateSettingCreateView,
    AnnualStateSettingsListView,
    AnnualStateSettingUpdateView,
)

__all__ = [
    "AnnualStateSettingCreateView",
    "AnnualStateSettingUpdateView",
    "AnnualStateSettingsListView",
    "InvoiceCreateView",
    "InvoiceDetailView",
    "InvoiceListView",
    "InvoiceUpdateView",
    "InvoiceYearlyListView",
    "QualiDeleteView",
    "QualiUpdateView",
    "SeasonActorListView",
    "SeasonAvailabilityUpdateView",
    "SeasonAvailabilityView",
    "SeasonCreateView",
    "SeasonDetailView",
    "SeasonErrorsListView",
    "SeasonExportView",
    "SeasonHelperListView",
    "SeasonListRedirect",
    "SeasonListView",
    "SeasonOrgaListView",
    "SeasonPersonalPlanningExportFeed",
    "SeasonPersonalPlanningExportView",
    "SeasonPlanningExportView",
    "SeasonPlanningView",
    "SeasonStaffChoiceUpdateView",
    "SeasonToOpenView",
    "SeasonAvailabilityReminderView",
    "SeasonToRunningView",
    "SeasonUpdateView",
    "SessionCreateView",
    "SessionDeleteView",
    "SessionDetailView",
    "SessionExportView",
    "SessionStaffChoiceView",
    "SessionUpdateView",
    "SessionsListView",
    "QualiCreateView",
]



# Generated by Django 1.11.4 on 2017-08-31 10:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("orga", "0005_auto_20170829_0959"),
    ]

    operations = [
        migrations.AlterField(
            model_name="organization",
            name="abbr",
            field=models.CharField(
                blank=True, max_length=16, verbose_name="Abréviation"
            ),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("orga", "0001_initial"),
        ("challenge", "0002_auto_20150828_1405"),
    ]

    operations = [
        migrations.AddField(
            model_name="session",
            name="organization",
            field=models.ForeignKey(
                related_name="sessions",
                blank=True,
                null=True,
                to="orga.Organization",
                on_delete=models.CASCADE,
            ),
        ),
    ]
